from mext.threads.locks import Lock
from mext.threads.sync import Condition

from collections import deque
from time import time
import sys, heapq

__all__ = ['Queue', 'PriorityQueue', 'LifoQueue', 'Empty']


class Empty(Exception):
    "Exception raised by Queue.get(block=0)."
    pass


class Queue(object):
    def __init__(self, maxsize=0):
        self.maxsize = maxsize or sys.maxint
        self._init(maxsize)
        self.lock = Lock()
        # threads waiting to get an item wait on this condition
        self._cond_has_items = Condition(self.lock)
        # threads waiting to put an item wait on this one
        self._cond_has_space = Condition(self.lock)
        # threads waiting on wait_tasks() are waiting on this one
        self._cond_tasks = Condition(self.lock)
        self.tasks = 0 # tasks in queue or being processed

    def task_done(self):
        with self._cond_tasks:
            if not self.tasks:
                raise ValueError('task_done() called too many times')
            self.tasks -= 1
            if not self.tasks:
                self._cond_tasks.notify_all()

    def wait_tasks(self):
        with self._cond_tasks:
            while self.tasks:
                self._cond_tasks.wait()

    def __len__(self):
        return len(self.queue)

    @property
    def is_full(self):
        return len(self.queue) >= self.maxsize

    def put(self, item, timeout='never'):
        with self._cond_has_space:
            if not _wait_timeout(
                self._cond_has_space,
                lambda: len(self.queue) < self.maxsize,
                timeout
            ):
                return False
            self._put(item)
            self.tasks += 1
            self._cond_has_items.notify()
            return True

    def get(self, timeout='never'):
        with self._cond_has_items:
            if not _wait_timeout(
                self._cond_has_items,
                lambda: len(self.queue),
                timeout
            ):
                raise Empty
            self._cond_has_space.notify()
            return self._get()

    def _init(self, maxsize):
        q = self.queue = deque()
        self._put = q.append
        self._get = q.popleft


class PriorityQueue(Queue):
    def _init(self, maxsize):
        self.queue = []

    def _put(self, item, heappush=heapq.heappush):
        heappush(self.queue, item)

    def _get(self, heappop=heapq.heappop):
        return heappop(self.queue)


class LifoQueue(Queue):
    def _init(self, maxsize):
        q = self.queue = []
        self._put = q.append
        self._get = q.pop



def _wait_timeout(cond, check, timeout):
    if timeout == 'never': # wait for as long as necessary
        while not check():
            cond.wait()
        return True
    elif not timeout: # nonblocking check
        return check()
    elif timeout < 0:
        raise ValueError("'timeout' must be a positive number")
    else:
        endtime = time() + timeout
        while not check():
            remaining = endtime - time()
            if remaining <= 0:
                return False
            cond.wait(remaining)
        return True

